"""Write a note for a paper."""

import os
import sys

import click

import zoia.config


@click.command()
@click.argument('citekey', required=True)
def note(citekey):
    """Write a note for a document."""
    metadata = zoia.metadata.load_metadata()
    if citekey not in metadata:
        click.secho(f'Citekey {citekey} does not exist in library.', fg='red')
        sys.exit(1)

    note_path = os.path.join(
        zoia.config.get_library_root(), citekey, 'notes.md'
    )

    if not os.path.isfile(note_path):
        # TODO: Add a header here.
        pass

    click.edit(filename=note_path)



import unittest
import unittest.mock

from .context import zoia
import zoia.citekey
import zoia.metadata


class TestHelpers(unittest.TestCase):
    def test__get_title_start(self):
        self.assertEqual(
            zoia.citekey._get_title_start(
                'Can quantum-mechanical description of physical reality be '
                'considered complete'
            ),
            'can',
        )

        self.assertEqual(
            zoia.citekey._get_title_start(
                'On the electrodynamics of moving bodies'
            ),
            'electrodynamics',
        )

    def test__apply_citekey_format(self):
        citekey = zoia.citekey._apply_citekey_format(
            name_string='Doe',
            year=99,
            first_word_of_title='foo',
            identifier=None,
        )
        self.assertEqual(citekey, 'doe99-foo')

        citekey = zoia.citekey._apply_citekey_format(
            name_string='Doe',
            year=99,
            first_word_of_title='foo',
            identifier='a',
        )
        self.assertEqual(citekey, 'doe99a-foo')

    def test__generate_identifiers(self):
        for identifier in zoia.citekey._generate_identifiers():
            self.assertEqual(identifier, 'b')
            break

        for i, identifier in enumerate(zoia.citekey._generate_identifiers()):
            if i == 0:
                self.assertEqual(identifier, 'b')
            elif i == 1:
                self.assertEqual(identifier, 'c')
            elif i == 24:
                self.assertEqual(identifier, 'z')
            elif i == 25:
                self.assertEqual(identifier, 'aa')
            elif i == 26:
                self.assertEqual(identifier, 'ab')
            elif i == 51:
                self.assertEqual(identifier, 'ba')
            elif i > 51:
                break


class TestCreateCitekey(unittest.TestCase):
    @unittest.mock.patch('zoia.citekey.zoia.metadata.load_metadata')
    def test_create_citekey_one_author_no_collision(self, mock_load_metadata):
        metadatum = zoia.metadata.Metadatum(
            title='The Foo Bar', authors=[['John', 'Doe']], year=1999
        )

        mock_load_metadata.return_value = {'doe00-baz': None}
        citekey = zoia.citekey.create_citekey(metadatum)
        self.assertEqual(citekey, 'doe99-foo')

    @unittest.mock.patch('zoia.citekey.zoia.metadata.load_metadata')
    def test_create_citekey_one_author_with_collision(
        self, mock_load_metadata
    ):
        metadatum = zoia.metadata.Metadatum(
            title='The Foo Bar', authors=[['John', 'Doe']], year=1999
        )

        mock_load_metadata.return_value = {'doe99-foo': None}
        citekey = zoia.citekey.create_citekey(metadatum)
        self.assertEqual(citekey, 'doe99b-foo')

    @unittest.mock.patch('zoia.citekey.zoia.metadata.load_metadata')
    def test_create_citekey_two_authors_no_collision(self, mock_load_metadata):
        metadatum = zoia.metadata.Metadatum(
            title='The Foo Bar',
            authors=[['John', 'Doe'], ['Jane', 'Roe']],
            year=1999,
        )

        mock_load_metadata.return_value = {'doe00-baz': None}
        citekey = zoia.citekey.create_citekey(metadatum)
        self.assertEqual(citekey, 'doe+roe99-foo')

    @unittest.mock.patch('zoia.citekey.zoia.metadata.load_metadata')
    def test_create_citekey_three_authors_no_collision(
        self, mock_load_metadata
    ):
        metadatum = zoia.metadata.Metadatum(
            title='The Foo Bar',
            authors=[['John', 'Doe'], ['Jane', 'Roe'], ['Joe', 'Bloggs']],
            year=1999,
        )

        mock_load_metadata.return_value = {'doe00-baz': None}
        citekey = zoia.citekey.create_citekey(metadatum)
        self.assertEqual(citekey, 'doe+99-foo')



"""Functionality to handle ISBNs."""

import isbnlib


def _isbn_has_valid_checksum(identifier):
    """Determine whether the given ISBN has a valid checksum."""
    if len(identifier) == 10:
        identifier = '978' + identifier

    numerals = [int(char) for char in identifier]
    checksum = 0
    for i, numeral in enumerate(numerals):
        weight = 1 if i % 2 == 0 else 3
        checksum += weight * numeral

    return (checksum % 10) == 0


def is_isbn(identifier):
    """Determine whether the identifier could be an ISBN."""

    identifier = normalize(identifier)
    return isbnlib.is_isbn10(identifier) or isbnlib.is_isbn13(identifier)


def normalize(identifier):
    """Remove a possible prefix from the identifier."""

    identifier = identifier.lower()
    identifier = identifier.replace('-', '')
    prefix = 'isbn:'
    if identifier.startswith(prefix):
        identifier = identifier[len(prefix) :]

    if isbnlib.is_isbn10(identifier):
        identifier = isbnlib.to_isbn13(identifier)

    return identifier



"""Open a document in the library."""

import os
import sys

import click

import zoia.config


def _open(citekey):
    document_path = os.path.join(
        zoia.config.get_library_root(), citekey, 'document.pdf'
    )
    if not os.path.isfile(document_path):
        raise FileNotFoundError(f'No document found for citekey {citekey}.')

    click.launch(document_path)


@click.command(name='open')
@click.argument('citekey', required=True)
def open_(citekey):
    """Open a document in the library."""
    try:
        _open(citekey)
    except FileNotFoundError as e:
        click.secho(str(e), fg='red')
        sys.exit(1)



"""Tools to interact with the library metadata."""

import json
import os
from dataclasses import dataclass
from typing import List

from zoia.config import get_library_root
from zoia.config import ZOIA_METADATA_FILENAME
from zoia.normalization import split_name


@dataclass
class Metadatum:
    title: str
    authors: List[str]
    year: int

    @classmethod
    def from_dict(cls, d):
        return cls(
            title=d['title'],
            authors=d['authors'],
            year=d['year'],
        )

    def __post_init__(self):
        if not isinstance(self.authors, list):
            raise TypeError(
                f'authors attribute must be a list, but got type '
                f'{type(self.authors)}.'
            )

        self.authors = [
            split_name(elem) if isinstance(elem, str) else elem
            for elem in self.authors
        ]

    def __str__(self):
        if len(self.authors) == 1:
            author_str = self.authors[0][1]
        elif len(self.authors) == 2:
            author_str = self.authors[0][1] + ' & ' + self.authors[1][1]
        else:
            author_str = self.authors[0][1] + ' et al.'

        s = f'{author_str} ({self.year}), '
        title_str = []
        str_len = len(s) + 2  # Quotation marks around the title add two chars.
        for i_word, word in enumerate(self.title.split()):
            str_len += len(word) + 1
            title_str.append(word)
            if str_len > 65 and i_word > 2:
                title_str.append('...')
                break
        title_str = ' '.join(title_str)
        return s + f'"{title_str}"'


def load_metadata():
    """Load the metadata for the library."""

    library_root = get_library_root()
    if library_root is None:
        return {}

    metadata_filename = os.path.join(library_root, ZOIA_METADATA_FILENAME)
    with open(metadata_filename) as fp:
        metadata = json.load(fp)

    return metadata


def _write_metadata(metadata):
    """Write the metadata for the library to disk.

    Note that this will overwrite any existing metadata.

    """
    library_root = get_library_root()
    if library_root is None:
        raise RuntimeError('No library root set.  Cannot write metadata!')

    metadata_filename = os.path.join(library_root, ZOIA_METADATA_FILENAME)
    with open(metadata_filename, 'w') as fp:
        json.dump(metadata, fp, indent=4, sort_keys=True)


def initialize_metadata():
    """Initialize an empty metadata file on the disk."""
    library_root = get_library_root()
    metadata_filename = os.path.join(library_root, ZOIA_METADATA_FILENAME)

    if os.path.exists(metadata_filename):
        raise RuntimeError(
            f'Metadata file already exists at {metadata_filename}'
        )

    _write_metadata({})


def append_metadata(key, value):
    """Append the given data to the metadata file."""
    metadata = load_metadata()
    if key in metadata:
        raise KeyError(f'Key {key} is already present.')

    metadata[key] = value
    _write_metadata(metadata)


def replace_metadata(key, value):
    """Replace the data for a given key."""
    metadata = load_metadata()
    if key not in metadata:
        raise KeyError(f'Key {key} not present.')

    metadata[key] = value
    _write_metadata(metadata)


def rename_key(old_key, new_key):
    """Rename a citekey in the metadata."""
    metadata = load_metadata()

    if new_key in metadata:
        raise KeyError(f'Key {new_key} is already present.')

    metadata[new_key] = metadata.pop(old_key)
    _write_metadata(metadata)


def get_arxiv_ids():
    """Return a set of all existing arXiv identifiers."""
    metadata = load_metadata()
    return {
        elem['arxiv_id'] for elem in metadata.values() if 'arxiv_id' in elem
    }


def get_isbns():
    """Return a set of all existing ISBNs."""
    metadata = load_metadata()
    return {elem['isbn'] for elem in metadata.values() if 'isbn' in elem}


def get_dois():
    """Return a set of all existing DOIs."""
    metadata = load_metadata()
    return {elem['doi'] for elem in metadata.values() if 'doi' in elem}



import unittest

from ..context import zoia
import zoia.ids.arxiv


class TestArxivValidIds(unittest.TestCase):
    def test__is_valid_old_style_arxiv_id(self):
        self.assertFalse(zoia.ids.arxiv._is_valid_old_style_arxiv_id('foo'))
        self.assertFalse(zoia.ids.arxiv._is_valid_old_style_arxiv_id('123'))
        self.assertTrue(
            zoia.ids.arxiv._is_valid_old_style_arxiv_id('astro-ph/9901001')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_old_style_arxiv_id('astro-ph/9913001')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_old_style_arxiv_id('2001.00001')
        )

    def test__is_valid_new_style_arxiv_id(self):
        self.assertFalse(zoia.ids.arxiv._is_valid_new_style_arxiv_id('foo'))
        self.assertFalse(zoia.ids.arxiv._is_valid_new_style_arxiv_id('123'))
        self.assertTrue(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('2001.00001')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('20.0100001')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('astro-ph/9901001')
        )
        self.assertTrue(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('2001.00001v2')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('2001.00001v2v')
        )
        self.assertFalse(
            zoia.ids.arxiv._is_valid_new_style_arxiv_id('2001.00001v2foo')
        )

    def test_is_arxiv(self):
        self.assertFalse(zoia.ids.arxiv.is_arxiv('foo'))
        self.assertFalse(zoia.ids.arxiv.is_arxiv('123'))
        self.assertTrue(zoia.ids.arxiv.is_arxiv('arXiv:2001.00001'))
        self.assertTrue(zoia.ids.arxiv.is_arxiv('arXiv:astro-ph/9901001'))


class TestNormalize(unittest.TestCase):
    def test_normalize(self):
        self.assertEqual(zoia.ids.arxiv.normalize('2001.00001'), '2001.00001')

        self.assertEqual(
            zoia.ids.arxiv.normalize('arXiv:2001.00001'), '2001.00001'
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('http://www.arxiv.org/abs/2001.00001'),
            '2001.00001',
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('https://www.arxiv.org/abs/2001.00001'),
            '2001.00001',
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('https://arxiv.org/abs/2001.00001'),
            '2001.00001',
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('https://arxiv.org/pdf/2001.00001'),
            '2001.00001',
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('https://arxiv.org/pdf/2001.00001.pdf'),
            '2001.00001',
        )

        self.assertEqual(
            zoia.ids.arxiv.normalize('arxiv.org/pdf/2001.00001.pdf'),
            '2001.00001',
        )



"""Entry-point to the zoia CLI."""

import click

from zoia.add import add
from zoia.edit import edit
from zoia.init import init
from zoia.note import note
from zoia.open import open_


@click.group()
@click.version_option()
def zoia():
    """The main entry point into `zoia`."""


zoia.add_command(add)
zoia.add_command(edit)
zoia.add_command(init)
zoia.add_command(note)
zoia.add_command(open_)

if __name__ == '__main__':
    zoia()



"""Module to check whether an identifier is a PDF."""

import os


def is_pdf(identifier):
    """Check whether an idenitfier is a PDF."""
    if not os.path.isfile(identifier):
        return False

    with open(identifier, 'rb') as fp:
        return fp.read(4) == b'%PDF'



import os
import tempfile
import unittest
import unittest.mock
from pathlib import Path

from click.testing import CliRunner

from .context import zoia
import zoia.init


class TestIsValidInitDir(unittest.TestCase):
    def test__is_valid_init_dir_none(self):
        self.assertFalse(zoia.init._is_valid_init_dir(None))

    def test__is_valid_init_dir_no_exists(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            directory = os.path.join(tmpdir, 'foo')
            self.assertTrue(zoia.init._is_valid_init_dir(directory))

    def test__is_valid_init_dir_exists_and_empty(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            self.assertTrue(zoia.init._is_valid_init_dir(tmpdir))

    def test__is_valid_init_dir_exists_not_empty(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            (Path(tmpdir) / 'foo').touch()
            self.assertFalse(zoia.init._is_valid_init_dir(tmpdir))


@unittest.mock.patch('zoia.init._is_valid_init_dir')
@unittest.mock.patch('zoia.init.os.getcwd')
class TestGetDefaultDirectory(unittest.TestCase):
    def test__get_default_directory_cwd(
        self, mock_getcwd, mock_is_valid_init_dir
    ):
        mock_getcwd.return_value = '/home/foo'
        mock_is_valid_init_dir.side_effect = lambda x: x == '/home/foo'
        observed_default_directory = zoia.init._get_default_directory()
        expected_default_directory = '/home/foo'
        self.assertEqual(
            observed_default_directory, expected_default_directory
        )

    def test__get_default_directory_subdir(
        self, mock_getcwd, mock_is_valid_init_dir
    ):
        mock_getcwd.return_value = '/home/foo'
        mock_is_valid_init_dir.side_effect = lambda x: x == '/home/foo/zoia'
        observed_default_directory = zoia.init._get_default_directory()
        expected_default_directory = '/home/foo/zoia'
        self.assertEqual(
            observed_default_directory, expected_default_directory
        )

    def test__get_default_directory_none(
        self, mock_getcwd, mock_is_valid_init_dir
    ):
        mock_getcwd.return_value = '/home/foo'
        mock_is_valid_init_dir.side_effect = lambda x: False
        observed_default_directory = zoia.init._get_default_directory()
        self.assertIsNone(observed_default_directory)


class TestInit(unittest.TestCase):
    @unittest.mock.patch('zoia.init.zoia.metadata.initialize_metadata')
    @unittest.mock.patch('zoia.init.set_library_root')
    @unittest.mock.patch('zoia.init.get_library_root')
    def test_init_denovo(
        self,
        mock_library_root,
        mock_set_library_root,
        mock_initialize_metadata,
    ):
        mock_library_root.return_value = None
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(zoia.init.init, input='\n')
            self.assertEqual(result.exit_code, 0)
            mock_set_library_root.assert_called_with(os.getcwd())
        mock_initialize_metadata.assert_called_once()



import os
import tempfile
import unittest
import unittest.mock

from .context import zoia
import zoia.config


class TestConfig(unittest.TestCase):
    @unittest.mock.patch('zoia.config.os.getenv')
    def test_get_config_filepath(self, mock_getenv):
        mock_getenv.return_value = '/home/foo/.config'
        observed_filepath = zoia.config.get_config_filepath()
        expected_filepath = '/home/foo/.config/zoia/config.yaml'
        self.assertEqual(observed_filepath, expected_filepath)

    @unittest.mock.patch('zoia.config.get_config_filepath')
    def test_load_config_empty(self, mock_config_filepath):
        with tempfile.TemporaryDirectory() as tmpdir:
            mock_config_filepath.return_value = os.path.join(
                tmpdir, 'zoia/config.yaml'
            )
            config = zoia.config.load_config()
        self.assertEqual(config, {})

    @unittest.mock.patch('zoia.config.get_config_filepath')
    def test_save_load_config(self, mock_config_filepath):
        config = {'directory': '/foo/bar'}
        with tempfile.TemporaryDirectory() as tmpdir:
            mock_config_filepath.return_value = os.path.join(
                tmpdir, 'zoia/config.yaml'
            )
            zoia.config.save_config(config)
            observed_config = zoia.config.load_config()
        self.assertEqual(observed_config, config)


class TestLibraryRoot(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.patcher = unittest.mock.patch('zoia.config.get_config_filepath')
        self.mock_config_filepath = self.patcher.start()
        self.mock_config_filepath.return_value = os.path.join(
            self.tmpdir.name, 'zoia/config.yaml'
        )

    def tearDown(self):
        self.tmpdir.cleanup()
        self.patcher.stop()

    def test_get_library_root_empty(self):
        library_root = zoia.config.get_library_root()
        self.assertIsNone(library_root)

    def test_set_get_library_root(self):
        zoia.config.set_library_root('foo')
        observed_library_root = zoia.config.get_library_root()
        self.assertEqual(observed_library_root, 'foo')



"""Module to classify identifiers."""

from enum import Enum

import zoia.ids


class IdType(Enum):
    DOI = 'doi'
    ARXIV = 'arxiv'
    ISBN = 'isbn'
    PDF = 'pdf'


class ZoiaUnknownIdentifierException(Exception):
    pass


def classify_and_normalize_identifier(identifier):
    """Classify an identifier and strip extraneous characters."""
    id_type = None
    if zoia.ids.pdf.is_pdf(identifier):
        id_type = IdType.PDF
    elif zoia.ids.arxiv.is_arxiv(identifier):
        id_type = IdType.ARXIV
        identifier = zoia.ids.arxiv.normalize(identifier)
    elif zoia.ids.isbn.is_isbn(identifier):
        id_type = IdType.ISBN
        identifier = zoia.ids.isbn.normalize(identifier)
    elif zoia.ids.doi.is_doi(identifier):
        id_type = IdType.DOI
        identifier = zoia.ids.doi.normalize(identifier)
    else:
        raise ZoiaUnknownIdentifierException(
            f'Cannot determine what kind of identifier is {identifier}.'
        )

    return id_type, identifier



import json
import os
import tempfile
import unittest
import unittest.mock
from pathlib import Path

import requests

from .context import zoia
import zoia.add


class TestValidateResponse(unittest.TestCase):
    def test__validate_response(self):
        response = requests.Response()
        response.status_code = 200
        self.assertTrue(zoia.add._validate_response(response, None))

    def test__validate_response_invalid(self):
        response = requests.Response()
        response.status_code = 404
        with self.assertRaises(zoia.add.ZoiaExternalApiException):
            self.assertFalse(zoia.add._validate_response(response, None))

        response.status_code = 300
        with self.assertRaises(zoia.add.ZoiaExternalApiException):
            self.assertFalse(zoia.add._validate_response(response, None))


class TestGetArxivMetadata(unittest.TestCase):
    @unittest.mock.patch('zoia.add.requests.get')
    def test__get_arxiv_metadata(self, mock_requests_get):
        response = unittest.mock.MagicMock()
        response.status_code = 200
        with open(
            os.path.join(
                os.path.dirname(__file__), 'fixtures/arxiv_response.xml'
            )
        ) as fp:
            response.text = fp.read()

        mock_requests_get.return_value = response

        observed_metadata = zoia.add._get_arxiv_metadata('1601.00001')
        expected_metadata = {
            'arxiv_id': '1601.00001',
            'authors': [['Michael', 'Kilgour'], ['Dvira', 'Segal']],
            'title': (
                'Inelastic effects in molecular transport junctions: The '
                'probe technique at high bias'
            ),
            'year': 2015,
            'month': 12,
            'doi': '10.1063/1.4944470',
        }

        self.assertEqual(observed_metadata, expected_metadata)


class TestGetDoiMetadata(unittest.TestCase):
    @unittest.mock.patch('zoia.add.requests.get')
    def test__get_doi_metadata(self, mock_requests_get):
        response = unittest.mock.MagicMock()
        response.status_code = 200
        with open(
            os.path.join(
                os.path.dirname(__file__), 'fixtures/doi_response.bib'
            )
        ) as fp:
            response.text = fp.read()

        mock_requests_get.return_value = response

        entry = zoia.add._get_doi_metadata('10.3847/1538-3881/aa9e09')
        self.assertEqual(entry['year'], 2018)
        self.assertEqual(
            entry['authors'],
            [['Christopher J.', 'Shallue'], ['Andrew', 'Vanderburg']],
        )


class TestGetIsbnMetadata(unittest.TestCase):
    @unittest.mock.patch('zoia.add.isbnlib.meta')
    def test__get_isbn_metadata(self, mock_meta):
        mock_meta.return_value = {
            'ISBN-13': '9781400848898',
            'Title': 'Modern Classical Physics',
            'Authors': ['Kip S. Thorne', 'Roger D. Blandford'],
            'Publisher': 'Princeton University Press',
            'Year': '2017',
            'Language': 'en',
        }

        observed_metadata = zoia.add._get_isbn_metadata('9781400848898')

        expected_metadata = {
            'isbn': '9781400848898',
            'title': 'Modern Classical Physics',
            'authors': [['Kip S.', 'Thorne'], ['Roger D.', 'Blandford']],
            'publisher': 'Princeton University Press',
            'year': 2017,
            'language': 'en',
        }

        self.assertEqual(observed_metadata, expected_metadata)


class TestAddArxivId(unittest.TestCase):
    @unittest.mock.patch('zoia.add.zoia.config.get_library_root')
    @unittest.mock.patch('zoia.add.zoia.metadata.get_arxiv_ids')
    @unittest.mock.patch('zoia.add.zoia.metadata.append_metadata')
    @unittest.mock.patch('zoia.add._get_doi_metadata')
    @unittest.mock.patch('zoia.add._get_arxiv_metadata')
    @unittest.mock.patch('zoia.add.requests.get')
    def test__add_arxiv_id(
        self,
        mock_requests_get,
        mock_get_arxiv_metadata,
        mock_get_doi_metadata,
        mock_append_metadata,
        mock_get_arxiv_ids,
        mock_get_library_root,
    ):
        response = unittest.mock.MagicMock()
        response.status_code = 200
        response.content = b'\xde\xad\xbe\xef'
        mock_requests_get.return_value = response

        mock_get_arxiv_ids.return_value = {'doe99-foo', 'smith01-bar'}

        mock_get_arxiv_metadata.return_value = {
            'arxiv_id': '1601.00001',
            'authors': [['Michael', 'Kilgour'], ['Dvira', 'Segal']],
            'title': (
                'Inelastic effects in molecular transport junctions: The '
                'probe technique at high bias'
            ),
            'year': 2015,
            'month': 12,
            'doi': '10.1063/1.4944470',
        }

        mock_get_doi_metadata.return_value = {
            'journal': 'The Journal of Chemical Physics',
            'title': (
                'Inelastic effects in molecular transport junctions: The '
                'probe technique at high bias'
            ),
            'authors': [['Michael', 'Kilgour'], ['Dvira', 'Segal']],
            'pages': '124107',
            'number': '12',
            'volume': '144',
            'publisher': '{AIP} Publishing',
            'month': 'mar',
            'year': 2016,
            'url': 'https://doi.org/10.1063%2F1.4944470',
            'doi': '10.1063/1.4944470',
            'ENTRYTYPE': 'article',
        }

        with tempfile.TemporaryDirectory() as tmpdir:
            mock_get_library_root.return_value = tmpdir

            zoia.add._add_arxiv_id('1601.00001')

            self.assertTrue(
                (
                    Path(tmpdir) / 'kilgour+segal16-inelastic/document.pdf'
                ).exists()
            )


class TestAddIsbn(unittest.TestCase):
    @unittest.mock.patch('zoia.add.zoia.config.get_library_root')
    @unittest.mock.patch('zoia.add.zoia.metadata.append_metadata')
    @unittest.mock.patch('zoia.add.zoia.citekey.create_citekey')
    @unittest.mock.patch('zoia.add._get_isbn_metadata')
    @unittest.mock.patch('zoia.add.zoia.metadata.get_isbns')
    def test__add_isbn(
        self,
        mock_get_isbns,
        mock_get_isbn_metadata,
        mock_create_citekey,
        mock_append_metadata,
        mock_get_library_root,
    ):
        mock_get_isbns.return_value = {}

        metadata = {
            'authors': [['Kip', 'Thorne'], ['Roger', 'Blandford']],
            'title': 'Modern Classical Physics',
            'year': '2017',
        }
        mock_get_isbn_metadata.return_value = metadata
        mock_citekey = 'thorne+blandford17-modern'
        mock_create_citekey.return_value = mock_citekey

        with tempfile.TemporaryDirectory() as tmpdir:
            mock_get_library_root.return_value = tmpdir
            zoia.add._add_isbn(identifier='9781400848898', citekey=None)

            self.assertTrue((Path(tmpdir) / mock_citekey).is_dir())

        mock_append_metadata.assert_called_once_with(mock_citekey, metadata)


class TestAddDoi(unittest.TestCase):
    @unittest.mock.patch('zoia.add.zoia.metadata.get_dois')
    @unittest.mock.patch('zoia.add.requests.get')
    @unittest.mock.patch('zoia.add._get_doi_metadata')
    @unittest.mock.patch('zoia.add.zoia.citekey.create_citekey')
    @unittest.mock.patch('zoia.add.zoia.config.get_library_root')
    @unittest.mock.patch('zoia.add.zoia.metadata.append_metadata')
    def test__add_doi(
        self,
        mock_append_metadata,
        mock_get_library_root,
        mock_create_citekey,
        mock__get_doi_metadata,
        mock_requests_get,
        mock_get_dois,
    ):
        mock_get_dois.return_value = {}

        mock_arxiv_response = unittest.mock.MagicMock()
        mock_arxiv_response.status_code = 200
        mock_arxiv_response.text = json.dumps({'arxivId': '1504.05957'})

        mock_pdf_response = unittest.mock.MagicMock()
        mock_pdf_response.status_code = 200
        mock_pdf_response.content = b'%PDF'

        def requests_side_effect(url):
            semantic_scholar_url = (
                'https://api.semanticscholar.org/v1/paper/10.1093/mnras/'
                'stv1552'
            )
            if url == semantic_scholar_url:
                return mock_arxiv_response
            elif url == 'https://arxiv.org/pdf/1504.05957.pdf':
                return mock_pdf_response
            else:
                raise ValueError(f'Bad url {url}')

        mock_requests_get.side_effect = requests_side_effect

        mock__get_doi_metadata.return_value = {
            'authors': [['J.', 'Antognini']],
            'title': 'Timescales of Kozai-Lidov oscillations',
            'year': 2015,
        }

        citekey = 'antognini15-timescales'
        mock_create_citekey.return_value = citekey

        with tempfile.TemporaryDirectory() as tmpdir:
            mock_get_library_root.return_value = tmpdir
            zoia.add._add_doi('10.1093/mnras/stv1552', citekey=None)

            self.assertTrue(
                (Path(tmpdir) / citekey / 'document.pdf').is_file()
            )

        mock_append_metadata.assert_called_once_with(
            'antognini15-timescales',
            {
                'authors': [['J.', 'Antognini']],
                'title': 'Timescales of Kozai-Lidov oscillations',
                'year': 2015,
                'arxiv_id': '1504.05957',
            },
        )



"""Add new documents to the library."""

import json
import os
import sys
from datetime import datetime
from multiprocessing.dummy import Process as ThreadProcess
from multiprocessing.dummy import Queue as ThreadQueue

import bibtexparser
import click
import feedparser
import isbnlib
import requests
from halo import Halo

import zoia.citekey
import zoia.config
import zoia.ids
import zoia.metadata
from zoia.ids.classification import classify_and_normalize_identifier
from zoia.ids.classification import IdType
from zoia.normalization import split_name


class ZoiaExternalApiException(Exception):
    pass


class ZoiaExistingItemException(Exception):
    pass


def _validate_response(response, identifier):
    if response.status_code == 200:
        return True
    elif response.status_code == 404:
        raise ZoiaExternalApiException(
            f'Identifier {identifier} does not appear to exist.'
        )
    elif response.status_code == 429:
        raise ZoiaExternalApiException(
            'Too many requests in too short a time. Please wait a few minutes '
            'before trying again.'
        )
    else:
        raise ZoiaExternalApiException(
            f'Error: Received HTTP status code {response.status_code}.',
        )


def _get_arxiv_metadata(identifier):
    """Get the DOI identifier (if it exists) from the arXiv API."""
    response = requests.get(
        f'https://export.arxiv.org/api/query?id_list={identifier}'
    )
    _validate_response(response, identifier)
    parsed_response = feedparser.parse(response.text)
    try:
        entry = parsed_response['entries'][0]
        if 'id' not in entry:
            raise ZoiaExternalApiException(
                f'Identifier {identifier} not found'
            )

    except (KeyError, IndexError):
        raise ZoiaExternalApiException(f'Identifier {identifier} not found')

    publication_date = datetime.strptime(
        entry['published'], '%Y-%m-%dT%H:%M:%SZ'
    )
    metadata = {
        'arxiv_id': identifier,
        'authors': [split_name(elem['name']) for elem in entry['authors']],
        'title': entry['title'].replace('\n ', ''),
        'year': publication_date.year,
        'month': publication_date.month,
    }

    if 'arxiv_doi' in entry:
        metadata['doi'] = entry['arxiv_doi']

    return metadata


def _get_doi_metadata(doi):
    response = requests.get(
        os.path.join('https://doi.org', doi),
        headers={'Accept': 'application/x-bibtex'},
    )
    _validate_response(response, doi)
    parser = bibtexparser.bparser.BibTexParser(
        customization=bibtexparser.customization.author
    )
    bib_db = bibtexparser.loads(response.text, parser=parser)
    entry = bib_db.entries[-1]
    if 'type' in entry:
        entry['type'] = entry.pop('ENTRYTYPE')

    if 'year' in entry:
        entry['year'] = int(entry['year'])

    entry['authors'] = entry.pop('author')
    entry['authors'] = [
        list(map(lambda x: x.strip(), reversed(elem.split(','))))
        for elem in entry['authors']
    ]

    del entry['ID']
    return entry


def _get_isbn_metadata(isbn):
    metadata = isbnlib.meta(isbn, service='goob')
    if not {'Authors', 'Title', 'Year'}.issubset(set(metadata)):
        raise ZoiaExternalApiException(
            f'Did not receive authors, title, or year for ISBN {isbn}.'
        )

    metadata['isbn'] = metadata.pop('ISBN-13')
    keys = list(metadata.keys())
    for key in keys:
        metadata[key.lower()] = metadata.pop(key)

    metadata['authors'] = list(map(split_name, metadata['authors']))

    try:
        metadata['year'] = int(metadata['year'])
    except ValueError:
        raise ZoiaExternalApiException(
            'Google Books returned a value "{}" for the year that could '
            'not be converted to an integer.'.format(metadata['year'])
        )

    return metadata


def _add_arxiv_id(identifier, citekey=None):
    if identifier in zoia.metadata.get_arxiv_ids():
        raise ZoiaExistingItemException(
            f'arXiv paper {identifier} already exists.'
        )

    # Downloading the PDF can take a while, so start it early in a separate
    # thread.
    pdf_queue = ThreadQueue()
    pdf_process = ThreadProcess(
        target=lambda q, x: q.put(requests.get(x)),
        args=(pdf_queue, f'https://arxiv.org/pdf/{identifier}.pdf'),
    )
    pdf_process.start()

    with Halo(text='Querying arXiv...', spinner='dots'):
        arxiv_metadata = _get_arxiv_metadata(identifier)

    if 'doi' in arxiv_metadata:
        with Halo(text='Querying DOI information...', spinner='dots'):
            arxiv_metadata.update(_get_doi_metadata(arxiv_metadata['doi']))

    if citekey is None:
        metadatum = zoia.metadata.Metadatum(
            authors=arxiv_metadata['authors'],
            title=arxiv_metadata['title'],
            year=arxiv_metadata['year'],
        )
        citekey = zoia.citekey.create_citekey(metadatum)
    paper_dir = os.path.join(zoia.config.get_library_root(), citekey)
    os.mkdir(paper_dir)

    zoia.metadata.append_metadata(citekey, arxiv_metadata)
    with Halo(text='Downloading PDF...', spinner='dots'):
        pdf = pdf_queue.get()
        pdf_process.join()

    if pdf.status_code == 200:
        with open(os.path.join(paper_dir, 'document.pdf'), 'wb') as fp:
            fp.write(pdf.content)
    else:
        click.secho('Was unable to fetch a PDF', fg='yellow')

    return metadatum


def _add_isbn(identifier, citekey):
    """Add an entry from an ISBN."""
    if identifier in zoia.metadata.get_isbns():
        raise ZoiaExistingItemException(f'ISBN {identifier} already exists.')

    with Halo(text='Querying ISBN metadata...', spinner='dots'):
        isbn_metadata = _get_isbn_metadata(identifier)

    if citekey is None:
        metadatum = zoia.metadata.Metadatum(
            authors=isbn_metadata['authors'],
            title=isbn_metadata['title'],
            year=isbn_metadata['year'],
        )
        citekey = zoia.citekey.create_citekey(metadatum)

    zoia.metadata.append_metadata(citekey, isbn_metadata)

    book_dir = os.path.join(zoia.config.get_library_root(), citekey)
    os.mkdir(book_dir)

    return metadatum


def _add_doi(identifier, citekey):
    """Add an entry from a DOI."""
    if identifier in zoia.metadata.get_dois():
        raise ZoiaExistingItemException(f'DOI {identifier} already exists.')

    # Query Semantic Scholar to get the corresponding arxiv ID (if there is
    # one) in a separate thread.
    arxiv_queue = ThreadQueue()
    arxiv_process = ThreadProcess(
        target=lambda q, x: q.put(requests.get(x)),
        args=(
            arxiv_queue,
            f'https://api.semanticscholar.org/v1/paper/{identifier}',
        ),
    )
    arxiv_process.start()

    with Halo(text='Querying DOI metadata...'):
        doi_metadata = _get_doi_metadata(identifier)

    metadatum = zoia.metadata.Metadatum(
        authors=doi_metadata['authors'],
        title=doi_metadata['title'],
        year=doi_metadata['year'],
    )

    if citekey is None:
        citekey = zoia.citekey.create_citekey(metadatum)

    paper_dir = os.path.join(zoia.config.get_library_root(), citekey)
    os.mkdir(paper_dir)

    with Halo(text='Querying Semantic Scholar for corresponding arXiv ID...'):
        arxiv_metadata_response = arxiv_queue.get()
        arxiv_process.join()

    arxiv_metadata = json.loads(arxiv_metadata_response.text)

    if (arxiv_id := arxiv_metadata.get('arxivId')) is not None:
        doi_metadata['arxiv_id'] = arxiv_id
        with Halo(text='Downloading PDF from arXiv...'):
            pdf_response = requests.get(
                f'https://arxiv.org/pdf/{arxiv_id}.pdf'
            )

        if pdf_response.status_code == 200:
            with open(os.path.join(paper_dir, 'document.pdf'), 'wb') as fp:
                fp.write(pdf_response.content)
        else:
            click.secho('Was unable to fetch a PDF', fg='yellow')

    zoia.metadata.append_metadata(citekey, doi_metadata)

    return metadatum


@click.command()
@click.argument('identifier', required=True)
@click.option(
    '--citekey',
    type=str,
    default=None,
    help='Specify the BibTex citation key.',
)
def add(identifier, citekey):
    if citekey is not None and citekey in zoia.metadata.load_metadata():
        click.secho(f'Citekey {citekey} already exists.', fg='red')
        sys.exit(1)

    try:
        id_type, normalized_identifier = classify_and_normalize_identifier(
            identifier
        )
    except zoia.ids.classification.ZoiaUnknownIdentifierException:
        click.secho(
            f'Cannot determine what kind of identifier {identifier} is.',
            fg='red',
        )
        sys.exit(1)

    try:
        if id_type == IdType.ARXIV:
            metadatum = _add_arxiv_id(normalized_identifier, citekey)
        elif id_type == IdType.ISBN:
            metadatum = _add_isbn(normalized_identifier, citekey)
        elif id_type == IdType.DOI:
            metadatum = _add_doi(normalized_identifier, citekey)
    except (ZoiaExternalApiException, ZoiaExistingItemException) as e:
        click.secho(f'{str(e)}', fg='red')
        sys.exit(1)

    # TODO: Add an ISBN
    # TODO: Add a PDF
    # TODO: Add a DOI
    # TODO: Add something manually

    # TODO: Add more to this success message.
    click.secho(f'Successfully added {str(metadatum)}.', fg='blue')



import tempfile
import unittest
import unittest.mock
from pathlib import Path

from .context import zoia
import zoia.metadata


@unittest.mock.patch('zoia.metadata.get_library_root')
class TestMetadata(unittest.TestCase):
    def setUp(self):
        self._tmpdir = tempfile.TemporaryDirectory()
        self.tmpdir = self._tmpdir.name

    def tearDown(self):
        self._tmpdir.cleanup()

    def test_write_load_metadata(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        zoia.metadata._write_metadata({'foo': 'bar'})
        metadata = zoia.metadata.load_metadata()
        self.assertEqual(metadata, {'foo': 'bar'})

    def test_append_metadata(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        zoia.metadata._write_metadata({'foo': 'bar'})
        zoia.metadata.append_metadata('baz', 'qux')
        metadata = zoia.metadata.load_metadata()
        self.assertEqual(metadata, {'foo': 'bar', 'baz': 'qux'})

    def test_initialize_metadata(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        self.assertFalse((Path(self.tmpdir) / '.metadata.json').exists())
        zoia.metadata.initialize_metadata()
        self.assertTrue((Path(self.tmpdir) / '.metadata.json').exists())

    def test_initialize_metadata_already_existing(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        (Path(self.tmpdir) / '.metadata.json').touch()
        with self.assertRaises(RuntimeError):
            zoia.metadata.initialize_metadata()

    def test_rename_key(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        zoia.metadata._write_metadata({'foo': 'bar', 'baz': 'qux'})
        zoia.metadata.rename_key('foo', 'quux')
        metadata = zoia.metadata.load_metadata()
        self.assertEqual(metadata, {'quux': 'bar', 'baz': 'qux'})

    def test_rename_key_existing_key(self, mock_get_library_root):
        mock_get_library_root.return_value = self.tmpdir
        zoia.metadata._write_metadata({'foo': 'bar', 'baz': 'qux'})
        with self.assertRaises(KeyError):
            zoia.metadata.rename_key('quuz', 'foo')

        with self.assertRaises(KeyError):
            zoia.metadata.rename_key('foo', 'baz')


class TestMetadatum(unittest.TestCase):
    def test_metadatum_from_dict(self):
        d = {
            'title': 'foo',
            'authors': ['John Doe', 'Jane Roe'],
            'year': 2001,
        }
        metadatum = zoia.metadata.Metadatum.from_dict(d)
        self.assertEqual(metadatum.title, 'foo')
        self.assertEqual(metadatum.authors, [['John', 'Doe'], ['Jane', 'Roe']])
        self.assertEqual(metadatum.year, 2001)

    def test_metadatum_str(self):
        metadatum = zoia.metadata.Metadatum(
            authors=['John Doe', 'Jane Roe'],
            year=2001,
            title='Foo',
        )

        self.assertEqual(str(metadatum), 'Doe & Roe (2001), "Foo"')


class TestGetArxivIds(unittest.TestCase):
    @unittest.mock.patch('zoia.metadata.load_metadata')
    def test_get_arxiv_ids(self, mock_load_metadata):
        mock_load_metadata.return_value = {
            'doe09-foo': {'arxiv_id': '0901.0123'},
            'roe19-baz': {'isbn': '9781499999990'},
            'smith10-bar': {'arxiv_id': '1002.1001'},
        }

        arxiv_ids = zoia.metadata.get_arxiv_ids()
        self.assertEqual(arxiv_ids, {'0901.0123', '1002.1001'})


class TestGetIsbns(unittest.TestCase):
    @unittest.mock.patch('zoia.metadata.load_metadata')
    def test_get_isbns(self, mock_load_metadata):
        mock_load_metadata.return_value = {
            'doe09-foo': {'arxiv_id': '0901.0123'},
            'roe19-baz': {'isbn': '9781499999990'},
            'smith10-bar': {'arxiv_id': '1002.1001'},
        }

        isbns = zoia.metadata.get_isbns()
        self.assertEqual(isbns, {'9781499999990'})


class TestGetDois(unittest.TestCase):
    @unittest.mock.patch('zoia.metadata.load_metadata')
    def test_get_dois(self, mock_load_metadata):
        mock_load_metadata.return_value = {
            'doe09-foo': {'arxiv_id': '0901.0123'},
            'roe19-baz': {'isbn': '9781499999990'},
            'smith10-bar': {'doi': '10.1000/foo'},
        }

        isbns = zoia.metadata.get_dois()
        self.assertEqual(isbns, {'10.1000/foo'})



import unittest

from ..context import zoia
import zoia.ids.doi


class TestDoi(unittest.TestCase):
    def test_is_doi(self):
        self.assertFalse(zoia.ids.doi.is_doi('foo'))
        self.assertFalse(zoia.ids.doi.is_doi('11.23915/distill.00005'))
        self.assertFalse(zoia.ids.doi.is_doi('10.1/distill.00005'))
        self.assertFalse(zoia.ids.doi.is_doi('10.23915'))
        self.assertFalse(zoia.ids.doi.is_doi('10.23915/'))
        self.assertTrue(zoia.ids.doi.is_doi('10.23915/distill.00005'))

    def test_normalize(self):
        self.assertEqual(
            zoia.ids.doi.normalize('10.23915/distill.00005'),
            '10.23915/distill.00005',
        )

        self.assertEqual(
            zoia.ids.doi.normalize('doi:10.23915/distill.00005'),
            '10.23915/distill.00005',
        )

        self.assertEqual(
            zoia.ids.doi.normalize('http://doi.org/10.23915/distill.00005'),
            '10.23915/distill.00005',
        )

        self.assertEqual(
            zoia.ids.doi.normalize(
                'https://dx.doi.org/10.23915/distill.00005'
            ),
            '10.23915/distill.00005',
        )

        self.assertEqual(
            zoia.ids.doi.normalize('https://arxiv.org/10.23915/distill.00005'),
            'https://arxiv.org/10.23915/distill.00005',
        )



import unittest

from ..context import zoia
import zoia.ids.isbn


class TestIsbn(unittest.TestCase):
    def test__isbn_has_valid_checksum_13_digit_isbn(self):
        self.assertTrue(
            zoia.ids.isbn._isbn_has_valid_checksum('9780691159027')
        )
        self.assertFalse(
            zoia.ids.isbn._isbn_has_valid_checksum('9780691159028')
        )

    def test__isbn_has_valid_checksum_10_digit_isbn(self):
        self.assertTrue(zoia.ids.isbn._isbn_has_valid_checksum('0716703440'))
        self.assertFalse(zoia.ids.isbn._isbn_has_valid_checksum('0716703441'))

    def test_is_isbn(self):
        self.assertFalse(zoia.ids.isbn.is_isbn('foo'))
        self.assertFalse(zoia.ids.isbn.is_isbn('012345'))
        self.assertFalse(zoia.ids.isbn.is_isbn('012345678910'))

        self.assertTrue(zoia.ids.isbn.is_isbn('0716703440'))
        self.assertTrue(zoia.ids.isbn.is_isbn('9780691159027'))

        self.assertFalse(zoia.ids.isbn.is_isbn('0716703441'))
        self.assertFalse(zoia.ids.isbn.is_isbn('9780691159028'))

        self.assertTrue(zoia.ids.isbn.is_isbn('0-7167-0344-0'))
        self.assertTrue(zoia.ids.isbn.is_isbn('978-0-69115-902-7'))

    def test_normalize(self):
        self.assertEqual(
            zoia.ids.isbn.normalize('0716703440'), '9780716703440'
        )
        self.assertEqual(
            zoia.ids.isbn.normalize('isbn:0716703440'), '9780716703440'
        )
        self.assertEqual(
            zoia.ids.isbn.normalize('arxiv:0716703440'), 'arxiv:0716703440'
        )



import unittest

from .context import zoia
import zoia.normalization


class TestNormalization(unittest.TestCase):
    def test_strip_diacritics(self):
        self.assertEqual(zoia.normalization.strip_diacritics('foo'), 'foo')
        self.assertEqual(zoia.normalization.strip_diacritics('Foo'), 'Foo')
        self.assertEqual(zoia.normalization.strip_diacritics('Fóò'), 'Foo')

    def test_normalize_string(self):
        self.assertEqual(zoia.normalization.normalize_name('foo'), 'foo')
        self.assertEqual(zoia.normalization.normalize_name('Foo'), 'foo')
        self.assertEqual(zoia.normalization.normalize_name('Fóò'), 'foo')

    def test_split_name(self):
        self.assertEqual(zoia.normalization.split_name('Doe'), ['', 'Doe'])
        self.assertEqual(
            zoia.normalization.split_name('John Doe'), ['John', 'Doe']
        )
        self.assertEqual(
            zoia.normalization.split_name('John van Doe'), ['John', 'van Doe']
        )
        self.assertEqual(
            zoia.normalization.split_name('John Q. Public'),
            ['John Q.', 'Public'],
        )

    def test_normalize_title_word(self):
        self.assertEqual(zoia.normalization.normalize_title_word('The'), 'the')
        self.assertEqual(
            zoia.normalization.normalize_title_word('"Why"'), 'why'
        )
        self.assertEqual(
            zoia.normalization.normalize_title_word(r'$\eta_3$'), 'eta3'
        )



"""Implementation of the `zoia init` command."""

import os
import sys
from pathlib import Path

import click

import zoia
import zoia.metadata
from zoia.config import get_library_root
from zoia.config import set_library_root


def _is_valid_init_dir(directory):
    """Determine whether a given directory is a valid root directory for zoia.

    A valid directory must either not exist or be empty.

    """
    if directory is None:
        return False

    return not Path(directory).exists() or len(os.listdir(directory)) == 0


def _get_default_directory():
    """Return a default directory for the zoia library root.

    The default library root will be set by trying the following directories in
    order until a valid directory is found:

    1. The user's current working directory.
    2. A subdirectory of the user's current working directory called `zoia`.

    If no valid directory is found from that list no default will be provided.

    """
    default_directories = [os.getcwd(), os.path.join(os.getcwd(), 'zoia')]
    for default_directory in default_directories:
        if _is_valid_init_dir(default_directory):
            return default_directory

    return None


@click.command()
@click.argument('directory', required=False, default=None)
def init(directory):
    """Initialize the `zoia` library."""
    current_library_root = get_library_root()
    if current_library_root is not None:
        click.secho(
            f'Warning: Found existing zoia library at {current_library_root}',
            fg='yellow',
        )
        confirmation = click.confirm('Choose a new directory?')
        if not confirmation:
            click.secho('Exiting.', fg='red')
            sys.exit(1)

    while not _is_valid_init_dir(directory):
        if directory is not None:
            click.secho(
                f'Error: Directory {directory} exists and is not empty.',
                fg='red',
            )
        directory = click.prompt(
            'Please provide a directory for your library',
            default=_get_default_directory(),
        )
        directory = os.path.expanduser(directory)

    os.makedirs(directory, exist_ok=True)
    set_library_root(directory)

    # Start with an empty dictionary in the metadata file.
    zoia.metadata.initialize_metadata()

    click.secho(
        f'Your zoia library was successfully initialized at {directory}!',
        fg='blue',
    )



import tempfile
import unittest
import unittest.mock
from pathlib import Path

from click.testing import CliRunner

from .context import zoia
import zoia.cli


class TestOpen(unittest.TestCase):
    @unittest.mock.patch('zoia.open.zoia.config.get_library_root')
    @unittest.mock.patch('zoia.open.click.launch')
    def test_open(self, mock_launch, mock_get_library_root):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmpdir:
            mock_get_library_root.return_value = tmpdir
            doc_dir = Path(tmpdir) / 'bar01-baz'
            doc_dir.mkdir()
            doc_path = doc_dir / 'document.pdf'
            doc_path.touch()

            result = runner.invoke(zoia.cli.zoia, ['open', 'bar01-baz'])
            self.assertEqual(result.exit_code, 0)



import tempfile
import unittest
import unittest.mock
from pathlib import Path

from click.testing import CliRunner

from .context import zoia
import zoia.cli


class TestNote(unittest.TestCase):
    @unittest.mock.patch('zoia.note.zoia.config.get_library_root')
    @unittest.mock.patch('zoia.note.zoia.metadata.load_metadata')
    @unittest.mock.patch('zoia.note.click.edit')
    def test_note_no_existing_note(
        self, mock_edit, mock_load_metadata, mock_get_library_root
    ):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmpdir:
            doc_dir = Path(tmpdir) / 'doe01-foo'
            doc_dir.mkdir()
            mock_load_metadata.return_value = {'doe01-foo': {}}
            mock_get_library_root.return_value = tmpdir

            result = runner.invoke(zoia.cli.zoia, ['note', 'doe01-foo'])
            self.assertEqual(result.exit_code, 0)



"""Edit the metadata for a document."""

import json
import sys

import click
import yaml

import zoia.config
from zoia.metadata import Metadatum


@click.command()
@click.option(
    '--syntax',
    type=click.Choice(['json', 'yaml'], case_sensitive=False),
    default='json',
    help='Which syntax to use to edit the metadata.',
)
@click.argument('citekey', required=True)
def edit(citekey, syntax):
    """Edit the metadata for a document."""
    metadata = zoia.metadata.load_metadata()
    if citekey not in metadata:
        click.secho(f'Citekey {citekey} does not exist in library.', fg='red')
        sys.exit(1)

    metadatum = metadata[citekey]
    text = None
    extension = None
    if syntax == 'json':
        text = json.dumps(metadatum, indent=4)
        extension = '.json'
    elif syntax == 'yaml':
        text = yaml.dump(metadatum, indent=4)
        extension = '.yaml'
    else:
        click.secho(f'Got unsupported syntax {syntax}.', fg='red')
        sys.exit(1)

    new_metadatum_str = click.edit(text=text, extension=extension)
    if new_metadatum_str is not None:
        if syntax == 'json':
            try:
                new_metadatum = json.loads(new_metadatum_str)
            except json.JSONDecodeError:
                # TODO: Add a better error message here and give the user the
                # opportunity to fix the mistake.
                click.secho(
                    'Could not parse JSON, didn\'t save metadata.', fg='red'
                )
                sys.exit(1)
        elif syntax == 'yaml':
            try:
                new_metadatum = yaml.safe_load(new_metadatum_str)
            except yaml.parser.ParserError:
                # TODO: Add a better error message here and give the user the
                # opportunity to fix the mistake.
                click.secho(
                    'Could not parse YAML, failed to save metadata.', fg='red'
                )
                sys.exit(1)
    else:
        click.secho('File not saved, not changing metadata.', fg='red')
        sys.exit(1)

    zoia.metadata.replace_metadata(citekey, new_metadatum)

    click.secho(
        f'Successfully edited metadata for '
        f'{str(Metadatum.from_dict(new_metadatum))}.',
        fg='blue',
    )



"""Module to manage configuration of `zoia`.

By default `zoia`'s configuration files will live in `$XDG_CONFIG_HOME/zoia/`.
If the user has not set `$XDG_CONFIG_HOME`, `zoia` will use the default
configuration directory of `$HOME/.config/zoia/`.

"""

import os

import yaml

ZOIA_METADATA_FILENAME = '.metadata.json'


def get_config_filepath():
    """Return the filepath of the configuration file."""
    default_config_root = os.path.join(os.path.expanduser('~'), '.config')
    config_root = os.getenv('XDG_CONFIG_HOME', default=default_config_root)
    return os.path.join(config_root, 'zoia/config.yaml')


def load_config():
    """Load the zoia configuration file if it exists."""
    config_filepath = get_config_filepath()
    if not os.path.exists(config_filepath):
        return {}

    with open(config_filepath) as fp:
        config = yaml.safe_load(fp)

    if not isinstance(config, dict):
        raise RuntimeError(
            f'Found existing configuration file at {config_filepath} but type '
            f'must be dict and found type {type(config)}.'
        )

    return config


def save_config(new_config):
    """Save the given configuration data."""
    if not isinstance(new_config, dict):
        raise TypeError(
            f'Given configuration must be a dictionary but got type '
            f'{type(new_config)}.'
        )

    old_config = load_config()
    merged_config = {**old_config, **new_config}
    config_filepath = get_config_filepath()
    zoia_config_root = os.path.dirname(config_filepath)
    os.makedirs(zoia_config_root, exist_ok=True)
    with open(config_filepath, 'w') as fp:
        yaml.safe_dump(merged_config, fp)


def get_library_root():
    """Find the root directory of the `zoia` library."""
    config = load_config()
    return config['directory'] if config and 'directory' in config else None


def set_library_root(library_root):
    """Set the root directory of the `zoia` library."""
    config = {'directory': library_root}
    save_config(config)



"""Module to handle identifiers."""

import zoia.ids.arxiv as arxiv  # noqa: F401
import zoia.ids.classification as classification  # noqa: F401
import zoia.ids.doi as doi  # noqa: F401
import zoia.ids.isbn as isbn  # noqa: F401
import zoia.ids.pdf as pdf  # noqa: F401



"""Utilities to normalize strings."""

import string
import unicodedata


def strip_diacritics(s):
    """Remove diacritics from the string."""
    return ''.join(
        [
            char
            for char in unicodedata.normalize('NFD', s)
            if unicodedata.category(char) != 'Mn'
        ]
    )


def normalize_name(s):
    """Remove diacritics and return a lower-case version of the string."""
    s = strip_diacritics(s)
    return s.lower()


def split_name(name):
    """Split a name into a list of a first and last name."""
    names = name.split()

    last_names = []
    if len(names) == 1:
        last_names.append(name)
    else:
        for elem in reversed(names[1:]):
            if '.' in elem:
                break
            else:
                last_names.append(elem)
    last_name = ' '.join(reversed(last_names))
    first_name = ' '.join(names[: len(names) - len(last_names)])
    return [first_name, last_name]


def normalize_title_word(word):
    """Normalize a word in a title."""
    word = strip_diacritics(word)
    word = word.lower()
    good_characters = string.ascii_lowercase + string.digits
    return ''.join(filter(lambda x: x in good_characters, word))



"""Module to check the validity of DOIs."""

import re


def is_doi(identifier):
    """Determine whether the given identifier has a valid DOI format."""
    if not isinstance(identifier, str):
        return False

    identifier = identifier.lower()
    identifier = identifier.split('/')
    if len(identifier) < 2:
        return False

    if not identifier[0] or not identifier[1]:
        return False

    prefix = identifier[0]

    prefix = prefix.split('.')
    if len(prefix) < 2:
        return False

    if any(map(lambda x: not x.isnumeric(), prefix)):
        return False

    if prefix[0] != '10':
        return False

    return int(prefix[1]) >= 1000


def normalize(identifier):
    """Normalize a DOI."""

    identifier = identifier.lower()

    prefix = 'doi:'
    if identifier.startswith(prefix):
        identifier = identifier[len(prefix) :]
    else:
        pattern = re.compile(r'^(https?://)?(www\.)?(dx\.)?doi\.org/')
        if pattern.match(identifier) is not None:
            identifier = pattern.split(identifier)[-1]

    return identifier



"""Functionality to interface with arXiv references."""

import re

ARXIV_FIELDS = {
    'astro-ph',
    'cond-mat',
    'gr-qc',
    'hep-ex',
    'hep-lat',
    'hep-ph',
    'hep-th',
    'math-ph',
    'nlin',
    'nucl-ex',
    'physics',
    'quant-ph',
    'math',
    'cs',
    'q-bio',
    'q-fin',
    'stat',
    'eess',
    'econ',
}


def _is_valid_old_style_arxiv_id(identifier):
    """Determine if the given identifiier is a valid old style arXiv ID."""
    split_identifier = identifier.split('/')
    if len(split_identifier) != 2:
        return False
    subject, identifier = split_identifier

    split_subject = subject.split('.')
    if len(split_subject) > 2:
        return False
    elif len(split_subject) == 2:
        archive, version = split_subject
        if not version.isnumeric():
            return False
    else:
        archive = split_subject[0]

    if archive not in ARXIV_FIELDS:
        return False

    split_identifier = identifier.split('v')
    if len(split_identifier) > 2:
        return False
    elif len(split_identifier) == 2:
        identifier, version = split_identifier
    else:
        identifier = split_identifier[0]

    if not identifier.isnumeric():
        return False

    if len(identifier) != 7:
        return False

    year = int(identifier[:2])
    month = int(identifier[2:4])

    if not (month >= 1 and month <= 12):
        return False

    # Old version arXiv IDs don't exist before 1991 or after March 2007.
    if (year < 91 and year > 7) or (year == 7 and month > 3):
        return False

    return True


def _is_valid_new_style_arxiv_id(identifier):
    """Determine if the given identifier is a valid new style arXiv ID."""

    split_identifier = identifier.split('v')
    if len(split_identifier) > 2:
        return False
    elif len(split_identifier) == 2:
        identifier, version = split_identifier
        if not version.isnumeric():
            return False
    else:
        identifier = split_identifier[0]

    split_identifier = identifier.split('.')
    if len(split_identifier) != 2:
        return False

    prefix, suffix = split_identifier
    if not prefix.isnumeric() or not suffix.isnumeric():
        return False

    if len(prefix) != 4 or len(suffix) not in {4, 5}:
        return False

    month = prefix[2:4]
    if int(month) > 12:
        return False

    return True


def is_arxiv(identifier):
    """Determine whether or not the given identifier is a valid arXiv ID."""

    identifier = normalize(identifier)
    if identifier.lower().startswith('arxiv:'):
        identifier = identifier[len('arxiv:') :]

    valid_old_style_arxiv_id = _is_valid_old_style_arxiv_id(identifier)
    valid_new_style_arxiv_id = _is_valid_new_style_arxiv_id(identifier)

    return valid_old_style_arxiv_id or valid_new_style_arxiv_id


def normalize(identifier):
    """Remove the 'arxiv:' prefix or URL if it exists."""

    identifier = identifier.lower()
    if identifier.startswith('arxiv:'):
        identifier = identifier[len('arxiv:') :]
    else:
        pattern = re.compile(r'^(https?://)?(www\.)?arxiv\.org/')
        if pattern.match(identifier) is not None:
            stripped_identifier = pattern.split(identifier)[-1]
            if stripped_identifier.startswith('abs/'):
                identifier = stripped_identifier[len('abs/') :].rstrip('/')
            elif stripped_identifier.startswith('pdf/'):
                identifier = stripped_identifier[len('pdf/') :]
                if identifier.endswith('.pdf'):
                    identifier = identifier[: -len('.pdf')]

    return identifier



"""Setup tooling for `zoia`."""

import setuptools

import versioneer

with open('README.md') as fp:
    long_description = fp.read()

setuptools.setup(
    name='zoia',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Command line tool to manage references.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/joe-antognini/zoia',
    author='Joseph Antognini',
    author_email='joe.antognini@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=[
        'bibtexparser>=1.2.0',
        'click>=7.1.2',
        'feedparser>=5.2.1',
        'halo>=0.0.30',
        'isbnlib>=3.10.3',
        'pyyaml>=5.3.1',
        'requests>=2.24.0',
    ],
    entry_points={
        'console_scripts': ['zoia=zoia.cli:zoia'],
    },
    include_package_data=True,
    setup_requires=['pytest-runner'],
    test_requires=['pytest'],
    python_requires='>=3.8',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
    ],
)



import os
import tempfile
import unittest
from pathlib import Path

from ..context import zoia
import zoia.ids.pdf


class TestPdf(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_is_pdf_no_file(self):
        self.assertFalse(
            zoia.ids.pdf.is_pdf(os.path.join(self.tmpdir.name, 'foo.pdf'))
        )

    def test_is_pdf_empty_file(self):
        (Path(self.tmpdir.name) / 'foo.pdf').touch()
        self.assertFalse(
            zoia.ids.pdf.is_pdf(os.path.join(self.tmpdir.name, 'foo.pdf'))
        )

    def test_is_pdf_good_pdf(self):
        with open(os.path.join(self.tmpdir.name, 'foo.pdf'), 'wb') as fp:
            fp.write(b'%PDF')

        self.assertTrue(
            zoia.ids.pdf.is_pdf(os.path.join(self.tmpdir.name, 'foo.pdf'))
        )



"""Functions to create a unique citekey."""

import re
import string

import zoia.metadata
import zoia.normalization

# TODO: Add functionality for other citekey styles.

# Ignore common words in the title when generating a citekey.
TITLE_WORD_BLACKLIST = {'a', 'an', 'are', 'is', 'of', 'on', 'the'}


def _get_title_start(title):
    """Get the first non-blacklisted word in the title."""
    title_words = re.split(' |-', title)
    normalized_words = map(
        zoia.normalization.normalize_title_word, title_words
    )
    for word in normalized_words:
        if word and word not in TITLE_WORD_BLACKLIST:
            return word

    for word in normalized_words:
        if word:
            return word

    return ''


def _apply_citekey_format(
    name_string, year, first_word_of_title, identifier=None
):
    """Create a citekey in the appropriate format."""
    citekey = (
        f'{name_string}{year}'
        f'{identifier if identifier else ""}-{first_word_of_title}'
    )
    return zoia.normalization.normalize_name(citekey)


def _generate_identifiers():
    i = 1
    while True:
        identifiers = []
        j = i
        while j >= 0:
            identifiers.append(string.ascii_lowercase[j % 26])
            j = j // 26 - 1
        identifiers = identifiers[::-1]
        i += 1
        yield ''.join(identifiers)


def create_citekey(metadatum):
    """Create a unique citekey for the object."""

    n_citekey_authors = 2 if len(metadatum.authors) == 2 else 1
    last_names = [
        elem[1].split() for elem in metadatum.authors[:n_citekey_authors]
    ]
    normalized_names = map('-'.join, last_names)
    name_string = '+'.join(normalized_names)
    if len(metadatum.authors) > 2:
        name_string += '+'
    year = metadatum.year % 100
    first_word_of_title = _get_title_start(metadatum.title)

    proposed_citekey = _apply_citekey_format(
        name_string, year, first_word_of_title
    )

    metadata = zoia.metadata.load_metadata()

    # TODO: Add a note to the README that this behavior is currently resolving
    # collisions by the order the reference was added to zoia, not by the
    # original date or title.  (This should be user-configurable.)
    if proposed_citekey in metadata:
        for identifier in _generate_identifiers():
            proposed_citekey = _apply_citekey_format(
                name_string, year, first_word_of_title, identifier
            )
            if proposed_citekey not in metadata:
                break

    return proposed_citekey



import unittest
import unittest.mock

import json
from copy import copy

from click.testing import CliRunner

from .context import zoia
import zoia.cli


class TestEdit(unittest.TestCase):
    @unittest.mock.patch('zoia.note.zoia.metadata.replace_metadata')
    @unittest.mock.patch('zoia.note.zoia.metadata.load_metadata')
    @unittest.mock.patch('zoia.note.click.edit')
    def test_edit(self, mock_edit, mock_load_metadata, mock_replace_metadata):

        orig_metadata = {
            'authors': [['John', 'Doe'], ['Jane', 'Roe']],
            'title': 'Foo',
            'year': 2001,
        }
        new_metadata = copy(orig_metadata)
        new_metadata['year'] = 2002

        mock_load_metadata.return_value = {'doe+roe01-foo': orig_metadata}
        mock_edit.return_value = json.dumps(new_metadata)

        runner = CliRunner()
        result = runner.invoke(
            zoia.cli.zoia, ['edit', 'doe+roe01-foo', '--syntax', 'json']
        )
        self.assertEqual(result.exit_code, 0)

        mock_replace_metadata.assert_called_once_with(
            'doe+roe01-foo', new_metadata
        )
